module Main(main) where


import Array hiding(bounds)
import Char
import IO
import List
import Maybe
import Monad
import System
import Time
import Directory

import Control.Concurrent
import Control.Exception
import Control.Monad.Monoid
import Data.IORef
import System.Console.GetOpt
import Data.Unique
import qualified Posix

import Charset
import ConfigFile 
import Curses
import EIO
import ErrorLog
import ExampleConf
import Filter
import Gale
import GenUtil
import Help
import KeyName
import Puff
import RSA(readSymbolicLink)
import Screen
import Signal
import Version
import Locale


data Env = Env {
    envAction :: IO (),
    envVerbose :: Int,
    envConfig :: (Maybe String),
    envJustArgs :: Bool,
    envErrorLog :: Maybe String,
    envNoPufflog :: Bool,
    envNoWritePufflog :: Bool
    }


----------------
-- Configuration
----------------

configBuilder "GALE_ID" = do
    n <- first [getEnv "LOGNAME", getEnv "USER", Posix.getLoginName]
    d <- configLookup "GALE_DOMAIN"
    case d of 
        Just d -> return [("<LOGIN>@$GALE_DOMAIN",("GALE_ID", n ++ "@" ++ d))]
        Nothing -> return []
configBuilder _ = return []
    
getGaleProxy :: IO [String]
getGaleProxy = do
    gps <- configLookup "GALE_PROXY"
    case gps of 
	Nothing -> do
	    v <- configLookup "GALE_DOMAIN"
            case v of 
                Just v -> return $ hostStrings v
                Nothing -> return []
	(Just x) -> return (words x)


getGaleId :: IO String
getGaleId = do
    v <- configLookup "GALE_ID"
    case v of 
        Just x -> return x
        Nothing -> return ""

getGaleDomain :: IO String
getGaleDomain = do
    v <- configLookup "GALE_DOMAIN"
    case v of 
        Just x -> return x
        Nothing -> return ""
    

{-

configBuilder "GALE_PROXY" = do
    gd <- configLookup "GALE_DOMAIN"
    case gd of 
        Just gd -> return [("$GALE_DOMAIN", ("GALE_PROXY", (unwords $ hostStrings gd)))]
        Nothing -> return []

getGaleDomain :: IO String
getGaleDomain = do
    gps <- configLookupList "GALE_DOMAIN"
    case gps of
	(x:_) -> return x
	[] -> fmap Posix.nodeName Posix.getSystemID 
    

getGaleProxy :: IO [String]
getGaleProxy = do
    gps <- configLookupList "GALE_PROXY"
    case gps of 
	[] -> do
	    v <- getGaleDomain
	    return $ hostStrings v
	(x:_) -> return (words x)

getGaleId :: IO String
getGaleId  = do
    gps <- configLookupList "GALE_ID"
    case gps of 
	(g:_) -> return g
	[] -> do
	    n <- first [getEnv "LOGNAME", getEnv "USER", Posix.getLoginName]
	    d <- getGaleDomain
	    return $ n ++ "@" ++ d
-}

getGaleAliases :: IO [(String,Category)]
getGaleAliases = do
    v <- galeFile "aliases/"
    fs <- handle (\_ -> return []) $ getDirectoryContents v
    let f fn = do
	fc <- first [readFile (v ++ fn), readSymbolicLink (v ++ fn)]
	return (fn, catParseNew fc)
    ks <- mapM (\x -> tryMost (f x)) fs
    return [x | (Right x) <- ks] 

getIdInstance = do
    pid <- Posix.getProcessID
    n <- liftM Posix.nodeName Posix.getSystemID
    return $ n ++ "/" ++ show pid



varsLookupElse :: [String] -> IO String -> IO String
varsLookupElse ss action = do
    v <- fmap concat $ mapM configLookupList ss
    case v of
        (x:_) -> return x
        _ -> action

galeFile :: String -> IO String
galeFile fn = do
    gd <- getGaleDir
    return (gd ++ fn)

getEditor = varsLookupElse ["VISUAL", "EDITOR"] (return "vi")

getTmpFile = do
    pid <- Posix.getProcessID
    u <- newUnique
    return ("/tmp/ginsu.puff." ++ show pid ++ "." ++ show (hashUnique u))


insertKeys ic s = mapM_ (\c -> writeChan ic (Left c)) (stringToKeys s)






configGood = do 
    gd <- configLookup "GALE_DOMAIN"
    case gd of 
        Just _ -> return ()
        Nothing -> putStrLn "GALE_DOMAIN is not set! either set $GALE_DOMAIN in the enviornment or set it in ~/.gale/conf or ~/.gale/ginsu.conf\n" >> doCheckConfig

doCheckConfig = do
    galeDir <- getGaleDir
    gp <- getGaleProxy
    gid <- configLookupElse "GALE_ID" "* Not Found *"
    galeDomain <- configLookupElse "GALE_DOMAIN" "* Not Found *"
    galeAliases <- getGaleAliases
    galeSubscribe <- configLookupElse "GALE_SUBSCRIBE" "* Not Set! *"


    mapM_ putStrLn $ buildTableLL [ 
	( "GALE_DIR ",  galeDir),
	( "GALE_DOMAIN ", galeDomain),
	( "GALE_ID ", gid),
	( "GALE_PROXY ", simpleQuote gp), 
	( "GALE_SUBSCRIBE ", galeSubscribe) ]
    when (galeAliases /= []) $ putStrLn $ "GALE_ALIASES \n" ++ 
	unlines ( map  (\(l,(c,d)) -> "  " ++ l ++ " -> " ++ c ++ "@" ++ d)  galeAliases) 

    let p = (galeDir ++ "/auth/private/")
	knames = [p ++ gid, p ++ gid ++ ".gpri"]
    gn <- (mapM (\fn -> (doesFileExist fn >>= \b -> if b then return (Just fn) else return Nothing)) knames)
    case msum gn of 
	Just fn -> putStrLn $ "Private key found in: " ++ fn
	Nothing -> putStrLn $ "Private key for '" ++ gid ++ "' not found in " ++ p
    putStrLn ""
    exitSuccess
    

    
main = do
    setLocaleAll ""
    galeDir <- getGaleDir
    configSetup $  mconcat [mapConfig ("GINSU_" ++) configEnv, configFile (galeDir  ++ "ginsu.config"), configEnv, configFile (galeDir ++ "conf"),configDefault (parseConfigFile exampleConf), configBuilder]

    (flags, acats) <- ginsuOpts

    configGood

    withErrorMessage bugMsg $ do

    cs <- configLookup "CHARSET"
    charsetSetup cs
    setErrorLogPutStr Charset.csHPutStr

    logname <- case envErrorLog flags of 
	Just x -> return x
	Nothing -> galeFile "ginsu.errorlog" 

    withErrorLog logname $ withStartEndEntrys "Ginsu" $ do
    case (envVerbose flags) of
	1 -> do 
	    setLogLevel LogInfo
	    putLog LogNotice $ "Verbosity Level: Info" 
	n | n > 1 -> do
	    setLogLevel LogDebug
	    putLog LogNotice $ "Verbosity Level: Debug" 
	_ -> return ()
    
    mainThd <- myThreadId
    Posix.installHandler Posix.sigINT (Posix.Catch (killThread mainThd)) Nothing
    Posix.installHandler Posix.sigPIPE Posix.Ignore Nothing

    
    galeDir <- getGaleDir
    gp <- getGaleProxy
    gid <- getGaleId
    galeDomain <- getGaleDomain
    galeAliases <- getGaleAliases

    putLog LogInfo $ "GALE_DIR " ++ galeDir
    putLog LogInfo $ "GALE_DOMAIN " ++ galeDomain
    putLog LogInfo $ "GALE_ID " ++ gid
    putLog LogInfo $ "GALE_PROXY " ++ simpleQuote gp
    putLog LogInfo $ "GALE_ALIASES \n" ++ 
	unlines ( map  (\(l,(c,d)) -> "  " ++ l ++ " -> " ++ c ++ "@" ++ d)  galeAliases) 


    Control.Exception.bracket_ initCurses endWin $ do
    withCursor CursorInvisible $ do
    -- withMouseEventMask [ButtonClicked 1] $ do

    ic <- newChan
    b <- configLookupBool "DISABLE_SIGWINCH"
    
    redraw_r <- newRenderContext

    case (b,cursesSigWinch) of 
	(False,Just s) -> Posix.installHandler s (Posix.Catch (resizeRenderContext redraw_r (writeChan ic (Left KeyResize)))) Nothing >> return ()
	_ -> return ()

    gs <- fmap (concat . map words) $ configLookupList "GALE_SUBSCRIBE"
    gps <- getGaleProxy
    let c =  gs ++ acats
	nc = map ("_gale.notice@" ++) $ snub (snds (map catParseNew c))
    
    gc <- newGaleContext gps (snub $ c ++ nc)
    doRender widgetEmpty
    pl <- forkIO $ puffLoop ic gc
    gl <- forkIO $ getchLoop ic
    yo <- newIORef (0::Int)
    plog <- galeFile "ginsu.pufflog"
    ops <- if envNoPufflog flags then return [] else 
	doRender (widgetCenter $  widgetText "Reading pufflog...") >> readPuffs plog
    ps <- newSVar ((reverse $ zip [1..] (reverse ops)))
    mark_set_r <- newIORef (listArray (0,9) (replicate 10 []))
    mapM_ (\n -> checkMark n >>= \f -> modifyIORef mark_set_r f) [1..9]
    next_r <- newIORef (length ops + 1)
    unless (envNoWritePufflog flags || envNoPufflog flags) $ (forkIO $ pufflogLoop ps 0) >> return ()

    (Just s) <- configLookup "ON_STARTUP" 
    insertKeys ic s 
    
    doRender (widgetCenter $  widgetText "Entering mainloop...") 
    mainLoop gc ic yo ps next_r mark_set_r redraw_r 
    destroyGaleContext gc
    killThread pl
    killThread gl
    ps <- liftM (map snd) $ readSVar ps
    unless (envNoPufflog flags || envNoWritePufflog flags) $ doRender (widgetCenter $ widgetText "Writing pufflog...") >>  writePufflog ps

pufflogLoop ps_r n = do
    threadDelay 30000000
    mv <- fmap (fmap fst . listToMaybe) $ readSVar ps_r
    case mv of
	Just n' | n' /= n -> do
	    ps <- liftM (map snd) $ readSVar ps_r
	    writePufflog ps
	    pufflogLoop ps_r n'
	_ -> pufflogLoop ps_r n

writePufflog ps = do
    plog <- galeFile "ginsu.pufflog"
    plsize <- fmap (read . fromJust) $ configLookup "PUFFLOG_SIZE" 
    withPrivateFiles $ writePuffs plog (take plsize ps)


checkMark n = do
    v <- configLookup ("MARK_" ++ show n)
    case v of 
	Nothing -> return id 
	Just z -> case parseFilter z of
	    Nothing -> putLog LogError ("Could not parse mark:\n  " ++ z) >> return id
	    Just y -> return (// [(n - 1,[y])])


puffLoop ic gc = repeatM_ $  galeNextPuff gc >>= \p -> (writeChan ic $ Right p)
getchLoop ic = repeatM_ (getCh >>= \x -> writeChan ic (Left x)) 


apHead f (x:xs) = f x:xs
apHead _ [] = []


expandAliases :: [Category] -> IO [Category]
expandAliases cats = do
    as <- getGaleAliases 
    gd <- getGaleDomain
    let ec x@(c,"") = de $ head $ [(tc ++ drop (length f) c,td) | (f,(tc,td)) <- as, f `isPrefixOf` c] ++ [x]
	ec x = x
	de (c,"") = (c,gd)
	de x = x
    return $ map ec cats 


statusBarWidget :: SVar String -> SVar Int -> SVar Int -> Widget
statusBarWidget svm svs svpc = widgetHorizontalBox False [(NoExpand,m), (Expand, widgetEmpty), (NoExpand,s), (NoExpand,widgetText "/"), (NoExpand, pc) ] where
    m = newSVarWidget svm widgetText
    s = newSVarWidget svs (widgetText . show) 
    pc = newSVarWidget svpc (widgetText . show)
    

presenceView :: [(String,(String,String))] -> Widget
presenceView pl = widgetText pt where
    pt = unlines $ buildTableLL (concatMap rt (gf pl))
    gf ts = map (\ts -> (fst (head ts), snds ts)) (groupBy (\(a,_) (b,_) -> a == b) ts)
    rt (a,bs) = [(a,concatInter ", " (snub (snds bs)))]

presenceViewUser :: String -> [(String,(String,String))] -> Widget
presenceViewUser user pl = widgetText pt where
    pt = user ++ ":\n" ++ indentLines 2  (unlines $ buildTableLL ([b|(a,b) <- pl, a == user]))

	
--spaceIt w = widgetVerticalBox False [sw,(NoExpand,wh), sw] where
--    sw = (NoExpand, widgetText " ")
--    wh = widgetHorizontalBox False [sw,(NoExpand,w), sw]
    
dialog s = widgetCenter $ widgetAttr [AttrBold] $ widgetSimpleFrame wb where
    wb = widgetText $ unlines [les, es, les]
    les = replicate (length es) ' '
    es = (" " ++ s ++ " " )

puffTemplate :: IO Puff
puffTemplate = do
    idText <- getIdInstance
    gi <- getGaleId
    t <- getClockTime
    u <- newUnique
    gn <- configLookup "GALE_NAME"
    n <- case gn of
	Just n -> return (("message/sender",FragmentText n):)
	Nothing -> return id
    let mid = show t ++ " " ++ idText ++ " " ++ show (hashUnique u)
    let sig = [Signed (Key gi [])]
    let ef = n [
	    ("message.id", FragmentText mid),
	    ("id/class",FragmentText fullName),
	    ("id/instance", FragmentText idText),
	    ("id/time",FragmentTime t)] 
    return emptyPuff {fragments = ef, signature = sig}

mainLoop gc ic yor psr next_r mark_set_r rc = do 
    let helpText = "Need Help? Press F1"
    hw <- helpWidget 
    cache_render_r <- newIORef ([],-1)
    cache_filter_r <- newIORef []

    message_r <- newSVar helpText
    selected_r <- newSVar 0
    ps <- readSVar psr
    puffcount_r <- newSVar (length ps)
    connect (svarSignal psr) $ writeSVar puffcount_r . length
    filter_r <- newSVar []

    presence_r <- newSVar []

    presence <- configLookupElse "GALE_PRESENCE" "in.perhaps"
    myPresence <- newSVar presence
    
    presenceWidget <- widgetScroll (newSVarWidget presence_r presenceView)

    ct <- getClockTime 
    userActionTime <- newSVar ct
    resizeSignal <- newSignal

    cSig <- newSignal 


    let sb = statusBarWidget message_r selected_r puffcount_r
    let fsw = newSVarWidget filter_r $ \fs -> 
	    widgetHorizontalBox False (intersperse (NoExpand, widgetText " ") $ map (\w -> (NoExpand,widgetAttr [AttrReverse] (widgetText w))) (reverse $ map showFilter fs)) 


    selPuff <- newCache $ do
	    ps <- readSVar psr
	    selected <- readSVar selected_r
	    return $ lookup selected ps 
    connect (signalChanged psr) $ \_ -> cacheInvalidate selPuff
    connect (signalChanged selected_r) $ \_ -> cacheInvalidate selPuff

    connect (svarSignal selected_r) $ \_ -> do
	sp <- readVal selPuff
	case sp of 
	    Nothing -> writeSVar message_r "none"
	    Just p -> writeSVar message_r $ getAuthor p ++ " via " ++ (maybe "unknown" id $ getFragmentString p "id/class")

    let	keyError s k = setMessage (s ++ ": " ++ keysToString [k] ++ " -- " ++ helpText)
	setMessage m = writeSVar message_r m 
	scrollPuffs x = modifyIORef yor (+ x) >> signal cSig ()
	gfs n = do
	    mark_set <- readIORef mark_set_r
	    modifyFilter (const (mark_set ! n))
	filter_thread = do
	    p <- readVal selPuff
	    case p of 
		Just (Puff {cats = [c]}) -> modifyFilter (FilterCategory c:)
		Just (Puff {cats = cs}) -> modifyFilter (FilterOr (map FilterCategory cs):)
		_ -> return ()
	showTS x = st "d" dayI ++ st "h" hourI ++ st "m" minI ++ show secI ++ "s" where
		(dayI, hourI) = divMod hourI' 24
		(hourI', minI) = divMod minI' 60
		(minI',secI) = divMod x 60
		st c 0 = "" 
		st c n = show n ++ c
	    
	addPuff p = do
	    case getFragmentString p "notice/presence" of 
		Nothing -> return ()
		Just pn -> do
		    let np@(a',(b',_)) = (getAuthor p, (maybe "unknown" id (getFragmentString p "id/instance"), pn))
		    modifySVar presence_r (\xs -> sort (np:[x| x@(a,(b,_)) <- xs, a /= a' || b /= b']))
	    case getFragmentString p "question.receipt" of
		Nothing -> return ()
		Just _ -> buildReciept p >>= sendPuff
	    case getFragmentString p "message/body" of
		Nothing -> return ()
		Just n -> when (not $ (all isSpace n) && (all (("_gale" `isPrefixOf`) . fst) (cats p))) $ do
		    v <- configLookupList "BEEP"
		    let f = FilterOr (catMaybes (map parseFilter v))
		    when (filterAp f p) Curses.beep
		    n <- readIORef next_r
		    modifyIORef next_r (+1)
		    modifySVar psr (\ps -> ((n,p):ps)) 
		    writeIORef cache_render_r ([], -1) 
		    select_perhaps select_next
		    touchRenderContext rc
	getRendered = do
	    (ys,xs) <- scrSize
	    (cr,xs') <- readIORef cache_render_r
	    if xs == xs' then return cr else do
	    --if False then return cr else do
		ps <- readSVar psr 
		fs <- readSVar filter_r
		let f p = all (`filterAp` p) fs
		let nps =  [(x,rPuff xs p)| x@(_,p) <- ps] 
		writeIORef cache_render_r ((nps, xs)) 
		writeIORef cache_filter_r [x| x@((_,p),_) <- nps, f p]
		return nps

	composePuff :: IO () -> [Category] -> [String] -> IO ()
	composePuff done cs kwds = do
	    p <- puffTemplate 
	    ep <- editPuff (p {cats = cs, fragments = fragments p ++ [ ("message.keyword",FragmentText k) | k <- kwds]})
	    case ep of 
		Nothing -> setMessage "Puff aborted."
		Just p -> do
		    pcw <- puffConfirm addPuff gc done p
		    setRenderWidget rc pcw
		    --sendPuff p
	getPresenceFrags = do
	    mp <- readSVar myPresence
	    (TOD ct _) <- getClockTime
	    uaT@(TOD uat _) <- readSVar userActionTime
	    (it,is,notIdle) <- return $ case ct - uat of 
		n | n < 240 -> (epoch," (not idle)",True)
		n -> (uaT," (" ++ (showTS n) ++ " idle)",False)
	    return ([
		    ("notice/presence",FragmentText (mp ++ is)), 
		    ("status.presence", FragmentText mp), 
		    ("status.idle", FragmentTime it)],notIdle)
	buildReciept p = do
	    np <- puffTemplate
	    r <- getFragmentString p "question.receipt"
	    gid <- getGaleId
	    let rid = maybe [] (\x -> [("receipt.id",FragmentText x)]) $ getFragmentString p "message.id"
	    (pf,_) <- getPresenceFrags
	    return $ np { cats = [catParseNew r], fragments = fragments np ++ pf ++ rid ++ [("answer.receipt",FragmentText gid)]}
	presenceLoopPollTime = 120000000
	presenceLoop = do
	    w <- newWatcher userActionTime
	    let loopIsIdle = do
		    b <- (watcherHasChanged w)
		    when b $ do
			_ <- readVal w
			notIdle <- sendPresence True
			threadDelay presenceLoopPollTime 
			if notIdle then loopNotIdle else loopIsIdle
		    threadDelay presenceLoopPollTime 
		    loopIsIdle
		loopNotIdle = do
		    notIdle <- sendPresence False
		    threadDelay presenceLoopPollTime 
		    if notIdle then loopNotIdle else readVal w >> loopIsIdle
	    sendPresence True
	    loopNotIdle 
	sendPresence sendIfNotIdle = do
	    (n,d) <- fmap catParseNew getGaleId
	    (pf,notIdle) <- getPresenceFrags
	    when (sendIfNotIdle || not notIdle) $ do
		p <- puffTemplate
		galeSendPuff gc $ p {cats = [("_gale.notice." ++ n,d)], fragments = fragments p ++ pf}
	    return notIdle
	gonePresencePuff = do
	    (n,d) <- fmap catParseNew getGaleId
	    p <- puffTemplate
	    let ef = [("notice/presence",FragmentText "out.gone"), ("notice.presence", FragmentText "out.gone")]
	    return $ p {cats = [("_gale.notice." ++ n,d)], fragments = fragments p ++ ef }
	sendPuff p = do
	    -- Expand Aliases
	    ncats <- expandAliases (cats p)
	    p <- return $ p {cats = ncats}
	    gs <- fmap (concat . map words) $ configLookupList "GALE_SUBSCRIBE"
	    if (or [c `subCategory` catParseNew g| c <- cats p, g <- gs] || isJust (getFragmentString p "answer.receipt")) then do
		    forkIO $ galeSendPuff gc p
		    setMessage "Puff being sent."
		else do
		    addPuff p
		    gid <- getGaleId
		    forkIO $ galeSendPuff gc (p {fragments = ("question.receipt", FragmentText gid):fragments p }) 
		    setMessage "Puff being sent (return receipt requested)"
	modifyFilter f = do
	    modifySVar filter_r f
	    fs <- readSVar filter_r
	    nps <- getRendered
	    let f p = all (`filterAp` p) fs
	    writeIORef cache_filter_r [x| x@((_,p),_) <- nps, f p]
	    select_perhaps select_next >> select_perhaps select_prev
	    center_puff
	getFiltered = do
	    getRendered
	    readIORef cache_filter_r
	buf_size = getFiltered >>= \ps -> return $ sum [x| (_,(x,_)) <- ps]
	center_puff = do
	    (ys,xs) <- scrSize
	    selected <- readSVar selected_r
	    ps <- liftM reverse getFiltered
	    let f t (((n,_),(h,_)):_) | n == selected = (t,t+h)
		f t ((_,(h,_)):rest) = f (t + h) rest
		f _ [] = (1000000,0)
	    let (mn,mx) = (f 0) ps
	    modifyIORef yor (min mn)
	    modifyIORef yor (max (mx - (ys - 1) ))
	scroll_end = do
	    (ys,xs) <- scrSize
	    bs <- buf_size
	    writeIORef yor (max (bs - (ys - 1)) 0)
	doRedraw (xorigin,yorigin) (xs,ys) = do 
	    bs <- buf_size
	    modifyIORef yor (min (max 0 (bs - (ys - 1))))
	    selected <- readSVar selected_r
	    ps <- getFiltered
	    yo <- readIORef yor
	    let rp _ [] =  return () 
		rp y (((i,_),(n,f)):rest) = do
		    when ((0,ys) `ovl` (y - yo, y - yo + n)) $ f stdScr (y - yo) (xs,ys) (i == selected)
		    rp (y + n) rest
	    attempt (rp 0 (reverse ps))
	select_perhaps a = do
	    ps <- getFiltered
	    n <- readSVar selected_r
	    unless (maybe False (const True) $ lookup n (fsts ps)) a
	select_prev = do
	    ps <- liftM fsts getFiltered
	    let f sel ((n,p):_) | n < sel = n 
		f sel (_:ps) = f sel ps
		f sel [] = sel
	    modifySVar selected_r (\s ->(f s ps))
	select_next = do
	    ps <- liftM fsts getFiltered
	    let f sel _ ((n,p):ps) | n > sel = f sel n ps
		f sel bg ((n,_):ps) | n > sel = f sel bg ps
		f _ bg _ = bg
	    modifySVar selected_r (\s ->(f s s ps))


    let fw = widgetVerticalBox False [(ExpandFill, mwidget), (NoExpand, widgetAttr [AttrBold] sb), (NoExpand, fsw)]
	mwidget = widgetEmpty { render = rnd, changedSignal = cSig, processKey = pk}
	rnd canvas = doRedraw (origin canvas) (bounds canvas)
	continue = return True
	rcontinue = signal cSig () >> return True
	pk x = case x of 
	    (KeyChar '\x12') -> reconnectGaleContext gc >> continue
	    (KeyChar 'q') -> 
		let pk x = case x of 
			(KeyChar 'y') -> return False
			(KeyChar 'Y') -> return False
			_ -> setRenderWidget rc fw >> return True 
		    in setRenderWidget rc (stackedWidgets [keyCatcherWidget pk (dialog "really quit y/n?"), fw]) >> return True
	    (KeyChar 'Q') -> return False
	    (KeyChar '\n') -> scrollPuffs 1 >> continue
	    (KeyChar '\b') -> scrollPuffs (-1) >> continue
	    (KeyChar '\x05') -> scrollPuffs 1  >> continue
	    (KeyChar '\x19') -> scrollPuffs (-1)  >> continue
	    (KeyChar '\EOT') -> scrSize >>= \(ys,_) ->  scrollPuffs (ys `div` 2) >> continue
	    (KeyChar '\NAK') -> scrSize >>= \(ys,_) ->  scrollPuffs (- (ys `div` 2)) >> continue
	    (KeyChar '\ACK') -> scrSize >>= \(ys,_) ->  scrollPuffs (ys) >> continue
	    (KeyChar '\STX') -> scrSize >>= \(ys,_) ->  scrollPuffs (-ys) >> continue
	    KeyNPage -> scrSize >>= \(ys,_) -> scrollPuffs (ys) >> continue
	    KeyPPage -> scrSize >>= \(ys,_) -> scrollPuffs (-ys) >> continue
	    (KeyChar 'E') -> do
		gc <- galeFile "ginsu.config"
		e <- getEditor
		withProgram $ System.system (e ++ " " ++ gc)
                reloadConfigFiles
		touchRenderContext rc
		return True
	    (KeyChar 'g') -> writeSVar selected_r 0 >> select_perhaps select_next >> writeIORef yor 0 >> rcontinue 
	    (KeyF 20) -> scroll_end >> return True
	    (KeyChar 'G') -> do
		ps <- readSVar psr
		case ps of 
		    ((n,_):_) -> writeSVar selected_r n >> select_perhaps select_prev >> scroll_end >> rcontinue 
		    _ -> continue
	    (KeyChar 'j') -> select_next >> center_puff >> rcontinue
	    (KeyChar 'k') -> select_prev >> center_puff >> rcontinue
	    KeyDown -> select_next >> center_puff >> rcontinue
	    KeyUp -> select_prev >> center_puff >> rcontinue
	    (KeyChar 'u') -> modifyFilter  (drop 1) >> rcontinue
	    (KeyChar 'U') -> modifyFilter (const []) >> rcontinue
	    (KeyChar '!') -> modifyFilter (apHead f) >> rcontinue where
		f (FilterNot x) = x
		f x = FilterNot x
	    (KeyChar 'x') -> modifyFilter f >> rcontinue where
		f (x:y:r) = y:x:r
		f x = x
	    (KeyChar 't') -> filter_thread >> rcontinue
	    (KeyChar 'T') -> filter_thread >> rcontinue
	    (KeyChar 'C') -> 
		let pk x = case x of
			    (KeyChar n) | n `elem` "123456789" -> do
                                mark_set <- readIORef mark_set_r
                                modifyFilter ((mark_set ! (ord n - ord '1')) ++)
			    key -> keyError "Invalid mark" key  
			>> do
			setRenderWidget rc fw
			return True  
		   in setRenderWidget rc (stackedWidgets [keyCatcherWidget pk  (dialog "recall which mark? [1-9]"), fw]) >> return True 
	    (KeyChar 'm') -> 
		let pk x = case x of
			    (KeyChar n) | n `elem` "123456789" -> do
				fs <- readSVar filter_r
				modifyIORef mark_set_r (// ([(ord n - ord '1',fs)])) 
				setMessage $ "mark " ++ [n] ++ " set."
			    key -> keyError "Invalid mark" key  
			>> do
			setRenderWidget rc fw
			return True  
		   in setRenderWidget rc (stackedWidgets [keyCatcherWidget pk  (dialog "set which mark? [1-9]"), fw]) >> return True 
	    (KeyChar 'a') -> do
		p <- readVal selPuff
		case p of 
		    Nothing -> continue
		    Just p -> case getSigner p of
			Just a -> do
			    modifyFilter (FilterAuthor a:)
			    rcontinue
			Nothing -> setMessage "no signed puff to get author from" >> continue
	    (KeyChar '/') -> do
		(ys,xs) <- scrSize
		v <- commandRead justGetKey stdScr (ys - 1) "Filter: " ""
		case v >>= parseFilter >>= \x -> return (modifyFilter (x:)) of
		    Just a -> a
		    Nothing -> setMessage "Invalid filter"
		rcontinue
	    (KeyChar 'c') -> do
		(ys,xs) <- scrSize
		v <- commandRead justGetKey stdScr (ys - 1) "Filter: " ""
		case v >>= parseFilter >>= \x -> return (modifyFilter (x:)) of
		    Just a -> a
		    Nothing -> setMessage "Invalid filter"
		rcontinue
	    (KeyChar '~') -> do
		(ys,xs) <- scrSize
		v <- commandRead justGetKey stdScr (ys - 1) "Filter: " "~"
		case v >>= parseFilter >>= \x -> return (modifyFilter (x:)) of
		    Just a -> a
		    Nothing -> setMessage "Invalid filter"
		rcontinue
	    (KeyChar 'd') -> do
		p <- readVal selPuff
		case p of 
		    Nothing -> setMessage "No puff selected" >> continue
		    Just p -> do
			pdw <- puffDetailsWidget p
			setRenderWidget rc (keyCatcherWidget (\_ -> setRenderWidget rc fw >> return True) pdw) >> continue
	    (KeyChar 'f') -> do
		p <- readVal selPuff
		case p of
		    Nothing -> setMessage "No puff selected" 
		    Just p -> do
			pkw <- fmap (concat . (map words)) $ configLookupList "PRESERVED_KEYWORDS" 
			let kws = filter (`elem` pkw) (getFragmentStrings p "message.keyword")
			composePuff (setRenderWidget rc fw) (cats p) kws
		continue
	    (KeyChar 'p') -> composePuff (setRenderWidget rc fw)[] [] >> rcontinue
	    (KeyChar 'N') -> do
		p <- readVal myPresence
		(ys,xs) <- scrSize
		v <- commandRead justGetKey stdScr (ys - 1) "Presence: " p
		case v of 
		    Just v -> writeSVar myPresence v >> setMessage ("Presence Updated: " ++ v)
		    Nothing -> setMessage "Presence unchanged."
		continue
	    (KeyChar 'r') -> do
		p <- readVal selPuff
		case p of
		    Nothing -> return ()
		    Just p -> do
			pkw <- fmap (concat . (map words)) $ configLookupList "PRESERVED_KEYWORDS" 
			let kws = filter (`elem` pkw) (getFragmentStrings p "message.keyword")
			composePuff(setRenderWidget rc fw) [catParseNew (getAuthor p)] kws 
		continue
	    (KeyChar n) | n `elem` "123456789" -> gfs (ord n - ord '1') >> rcontinue
	    key -> keyError "Invalid key" key >> rcontinue
	puffDetailsWidget p = do
	    let pv = newSVarWidget presence_r (presenceViewUser (getAuthor p))
	    v <- widgetScroll (widgetVerticalBox False [(NoExpand,widgetText $ showPuff p), (NoExpand, widgetText "\n\n\n"), (NoExpand, pv)])
	    return v
	justGetKey = fmapLeft keyCanon (readChan ic) >>= \v -> case v of
	    (Right p) -> do
		addPuff p >> justGetKey
	    (Left k) -> do
		    getClockTime >>= writeSVar userActionTime
		    return k
	nextKey = do
	    ce <- isEmptyChan ic
	    when ce $ tryDrawRenderContext rc
	    v <- fmapLeft keyCanon $ readChan ic
	    case v of 
		(Right p) -> do
		    addPuff p
		    s <- configLookupElse "ON_INCOMING_PUFF" ""
		    mapM_ (processKey mwidget) (stringToKeys s)
		    nextKey
		(Left (KeyChar '\f')) -> attempt (touchWin stdScr)  >> resizeRenderContext rc (return ()) >> nextKey
		(Left KeyResize) -> signal resizeSignal () >> nextKey
		(Left (KeyF 1)) -> setRenderWidget rc (keyCatcherWidget (\_ -> setRenderWidget rc fw >> return True) hw) >> nextKey
		(Left (KeyF 2)) -> setRenderWidget rc fw >> nextKey
		(Left (KeyF 3)) -> setRenderWidget rc (keyCatcherWidget (\_ -> setRenderWidget rc fw >> return True) presenceWidget) >> nextKey
		(Left k) -> do
		    getClockTime >>= writeSVar userActionTime
		    b <- keyRenderContext rc k 
		    if b then nextKey else return k
    gonePresencePuff >>= galeWillPuff gc 
    ploopID <- forkIO presenceLoop
    connect (signalChanged myPresence) (\_ -> sendPresence True >> return ())
    setRenderWidget rc fw
    nextKey
    killThread ploopID


-- makeURLRegex :: IO Regex 
-- makeURLRegex = regcomp "^(([^:/?#]+):)?(//([^/?#]*))?([^?#]*)(\\?([^#]*))?(#(.*))?" regExtended


rPuff :: Int -> Puff -> (Int, Window -> Int -> (Int,Int) -> Bool -> IO ())
rPuff mw p@(Puff {cats =cats, fragments = frags}) = (n + 2,f) where
    reformat s | length s < mw = [s]
    reformat s = let (a,b) = splitAt mw s in a : reformat b
    sender = maybe "" id $ getFragmentString p "message/sender" 
    body = maybe [] lines $ getFragmentString frags  "message/body" 
    time = getFragmentTime frags "id/time" `mplus` getFragmentTime frags "_ginsu.timestamp"
    kwds = getFragmentStrings p "message.keyword"
    body' = concatMap reformat body
    n = length body'
    addCat w (x,y) = wAttrOn w attrBold >> waddstr w x >> wAttrOff w attrBold >> waddstr w ('@':y) 
    f w y (xs,ys) selected = do
	when (y >= 0 && y < ys) $ do
	    wmove w y 0
	    --wAttrOn w attrBold
	    --waddstr w (fmtS "%-20.20s" sender)
	    --wAttrOff w attrBold
	    sequence (intersperse (space w) $ map (addCat w) cats)
	    space w >> space w
	    waddstr w (unwords (map ('/':) kwds))
	    st <- case time of 
		Just t -> liftM ft $ toCalendarTime t
		Nothing -> return ""
	    let lst = length st
	    wAttrOn w attrBold
	    mvwaddstr w y (mw - length sender - lst - 1) sender
	    wAttrOff w attrBold
	    when selected $ void standout
	    mvwaddstr w y (mw - lst) st
	    when selected $ void standend
	mapM_ (\(p,s) ->  when (y + p < ys) (mvwaddstr stdScr (y + p) 0 s)) (zip [1..] $ body')
	wAttrOn w attrBold
	let bs = (if selected then "======<" else "------<") ++ getAuthor p ++ ">" ++ cycle (if selected then "=" else "-")
	--when (y + 1 + n < ys) $ mvwaddstr w (y + 1 + n) 0 (take mw $ cycle $ if selected then "=" else  "-")
	when (y + 1 + n < ys) $ mvwaddstr w (y + 1 + n) 0 (take mw bs)
	wAttrOff w attrBold
	return ()
    ft (CalendarTime {ctHour = h, ctMin = m, ctSec = s}) = fn h ++ ":" ++ fn m ++ ":" ++ fn s
    fn n = let x = show n ; lx = length x in if lx < 2 then replicate (2 - lx) '0' ++ x else x

	
(a,b) `ovl` (x,y) | y < a = False
(a,b) `ovl` (x,y) | b < x = False
_ `ovl` _ = True


------------------
-- Curses routines
------------------

waddstr w s = Control.Exception.try (wAddStr w s) >> return ()
mvwaddstr w y x s = Control.Exception.try (mvWAddStr w y x s) >> return ()
wmove w y x = Control.Exception.try (wMove w y x) >> return ()

space w = waddstr w " "

--------------
-- help screen
--------------


helpWidget = do
    let w = widgetText helpTable
	s = "Help Screen"
	v = "* scroll with direction keys - F2 to return to main screen *"
    sw <- widgetScroll w 
    return $ widgetVerticalBox False [(NoExpand,widgetAttr [AttrBold] (widgetText s)),(NoExpand,widgetText v),(ExpandFill, widgetSimpleFrame sw)]

-------------------
-- Puff composition
-------------------

withPrivateFiles action = do
    om <- Posix.setFileCreationMask (Posix.groupModes `Posix.unionFileModes` Posix.otherModes)
    v <- action 
    Posix.setFileCreationMask om
    return v

noBodyWords fl = [x|x@(n,_) <- fl, n /= "message/body", n /= "message.keyword"]

editPuff :: Puff -> IO (Maybe Puff)
editPuff puff = do
    e <- getEditor
    fn <- getTmpFile
    eo <- configLookupList "EDITOR_OPTION"
    let it = ["To: " ++ simpleQuote (showDestination (cats puff) (getFragmentStrings puff "message.keyword")) , "--------"] 
    let mb = case getFragmentString puff "message/body" of
	    Just mb -> mb
	    Nothing -> "\n"
    withPrivateFiles $ writeRawFile fn (stringToBytes $ unlines it ++ mb)
    withProgram $ System.system (e ++ " " ++ unwords eo ++ " " ++ fn)
    pn <- fmap (lines . bytesToString)$ readRawFile fn 
    handleMost (\_ -> return ()) (removeFile fn) 
    if not (length pn > 1 && pn /= it) then return Nothing else do 
	let (cs',kwds') = readDestination (drop 4 (head pn))
	ncats <- expandAliases (cs')
        tw <- configLookupBool "TRIM_BLANKLINES"
        body <- return $ if not tw then (unlines $ (drop 2 pn)) else
            trimBlankLines (unlines $ (drop 2 pn))
        return $ if body == "" then Nothing else 
                Just puff { cats = ncats, fragments = noBodyWords (fragments puff) ++ [("message/body",FragmentText body)] ++ [ ("message.keyword",FragmentText k) | k <- kwds']}
    
showDestination cs kwds = (map catShowNew cs ++ map ('/':) kwds)
readDestination :: String -> ([Category],[String])
readDestination s = splitEither (map pi (simpleUnquote s)) where
    pi ('/':s) = Right s
    pi x = Left (catParseNew x)

sendPuff addPuff gc p = do
    -- Don't Expand Aliases, editPuff does
    --ncats <- expandAliases (cats p)
    --p <- return $ p {cats = ncats}
    gs <- fmap (concat . map words) $ configLookupList "GALE_SUBSCRIBE"
    if (or [c `subCategory` catParseNew g| c <- cats p, g <- gs] || isJust (getFragmentString p "answer.receipt")) then do
	    return ()
	else do
	    addPuff p
    forkIO $ galeSendPuff gc p

--------------------
-- Puff confirmation
--------------------

anonymousFragments = [("id/class",FragmentText "anonymous"), ("id/instance", FragmentText "anonymous"), ("message/sender", FragmentText "Anonymous")]

minusFrag fl s = [x |x@(n,_) <- fl, n /= s]

prettyPuff puff = unlines xs ++ body where
    to = ["To: " ++ simpleQuote (showDestination (cats puff) (getFragmentStrings puff "message.keyword")) ] 
    from = ["From: " ++ t| ("message/sender",FragmentText t) <- fragments puff]
    body = case [t | ("message/body",FragmentText t) <- fragments puff] of
	(t:_) -> t
	[] -> ""
    rr = if hasFragment puff "question.receipt" then ["Return Receipt: Yes"] else []
    xs = to ++ from ++ rr ++ ["-------"]
	 

puffConfirm :: (Puff -> IO ()) -> GaleContext -> IO () -> Puff -> IO Widget
puffConfirm addPuff gc done puff = do
    sbsv <- newSVar ""
    psv <- newSVar puff
    psvs <- newSVar (prettyPuff, showPuff)
    csv <- combineSVar psv psvs
    let pw = newSVarWidget csv (\(p,(f,_)) -> widgetText (f p))
    let w = (widgetCenter $  widgetText "SEND PUFF?")
	h = (widgetCenter $ widgetText "y:send q:abort e:edit r:returnRecipet h:allHeaders A:anonymize")
	pk (KeyChar 'q')  = done >> return True
	pk (KeyChar 'n')  = done >> return True
	pk (KeyChar 'r') = do
	    p <- readSVar psv
	    gid <- getGaleId
	    if hasFragment p "question.receipt" then 
		writeSVar psv (p {fragments = [x | x@(n,_) <- fragments p , n /= "question.receipt"]})
		  else
		    writeSVar psv (p {fragments = ("question.receipt", FragmentText gid) : fragments p})
	    return True
	pk (KeyChar 'h') = do
	    modifySVar psvs (\(x,y) -> (y,x))
	    return True
	pk (KeyChar 'A') = do
	    p <- readSVar psv
	    if hasFragment p "message.id" then do
		writeSVar psv (p {signature = [], fragments = anonymousFragments `mergeFrags` (fragments p `minusFrag` "message.id")})
	      else do
		pt <- puffTemplate
		writeSVar psv (p {signature = signature pt, fragments = (fragments pt `minusFrag` "id/time") `mergeFrags` fragments p})
	    return True
	pk (KeyChar 'y')  = do
	    puff <- readSVar psv
	    sendPuff addPuff gc puff 
	    done
	    return True
	pk (KeyChar 'e') = do
	    puff <- readSVar psv
	    np <- editPuff puff
	    case np of
		Just np -> writeSVar psv np
		Nothing -> return ()
	    return True
	pk key = do
	    writeSVar sbsv $  "Invalid keystroke: " ++ keysToString [key]
	    return True
    pb <- widgetScroll $ pw
    return $ keyCatcherWidget pk $ widgetVerticalBox False [(NoExpand, w), (NoExpand, widgetAttr [AttrBold] h), (ExpandFill, widgetSimpleFrame pb),(NoExpand, simpleStatusBarWidget sbsv)]

simpleStatusBarWidget :: SVar String -> Widget
simpleStatusBarWidget svm  = widgetAttr [AttrBold] $ newSVarWidget svm widgetText
    

-------------------------
-- options/initialization
-------------------------

env = Env { 
    envAction = return (),
    envVerbose = 0, 
    envConfig = Nothing,
    envJustArgs = False,
    envNoPufflog = False,
    envNoWritePufflog = False,
    envErrorLog = Nothing
    } 


options :: [OptDescr (Env -> Env)]
options = [
	Option ['v']  ["verbose"] (NoArg (\e->e{envVerbose = envVerbose e + 1})) "verbose output to errorlog, -vv for debugging info too.",
	Option ['V']  ["version"] (NoArg (doAction doVersion)) "print version information",
--	Option ['c']  ["config"] (ReqArg (\x e->e{envConfig = (Just x)}) "FILE") "use this configuration file",
	Option ['s']  ["sample"] (NoArg (doAction doSample)) "print sample configuration file to stdout",
	Option ['m']  ["man"] (NoArg (doAction doMan)) "print all internal help screens to stdout",
	Option ['e']  ["justargs"] (NoArg (\e->e{envJustArgs = True})) "only subscribe to command line arguments",
	Option ['P']  [] (NoArg (\e->e{envNoWritePufflog = True})) "do not write to pufflog",
	Option []  ["nopufflog"] (NoArg (\e->e{envNoPufflog = True})) "do not read or write pufflog",
	Option []  ["errorlog"] (ReqArg (\x e->e{envErrorLog = Just x}) "FILE") "log errors to file",
	Option []  ["testCurses"] (NoArg (doAction (cursesTest >> exitSuccess))) "print curses diagnostics",
	Option []  ["dumpKey"] (ReqArg (\x -> doAction (dumpKey x >> exitSuccess)) "KEYFILE") "print info for keyfile",
	Option []  ["checkconfig"] (NoArg (doAction doCheckConfig)) "check and print out configuration"
    ]

void a = a >> return ()

ginsuOpts :: IO (Env,[String])
ginsuOpts = do
    args <- getArgs
    r@(env,_) <- case (getOpt Permute options args) of
	   (as,n,[]) -> return (foldr ($) env as ,n)
    	   (_,_,errs) -> putErrDie (concat errs ++ usage)
    envAction env
    --handle (\_ -> putErrDie ("Could not find configuration file: ~/.gale/ginsu.config\n" ++ usageTrailer)) (loadStdConfiguration ".gale/ginsu.config")
    return r

doVersion = putStrLn fullName >> exitSuccess
doSample = putStr exampleConf >> exitSuccess
doMan = putStrLn usage >> putStrLn helpTable >> exitSuccess

usage = usageInfo usageHeader options ++ usageTrailer

doAction :: IO a -> (Env -> Env)
doAction a = \e -> e {envAction = a >> envAction e}



commandRead :: IO Curses.Key -> Window -> Int -> String -> String -> IO (Maybe String)
commandRead ic win yloc prompt init = withCursor CursorVisible  $ cr (reverse init) 
    where
    cr v = pc v >> ic >>= \x -> case x of
	(KeyChar '\n') -> return $ Just (reverse v)
	(KeyChar '\b') -> cr (drop 1 v)
	(KeyChar c) -> cr (c:v)
	_ -> return Nothing
    pc v = do
	wmove win yloc 0
	wClrToEol win
	wAttrOn win attrBold
	waddstr win prompt 
	wAttrOff win attrBold
	waddstr win (reverse v)
	wmove win yloc (length v + length prompt)
	refresh

--
-- graveyard
--
--	composePuff' cs kwds = do
--	    e <- getEditor
--	    fn <- getTmpFile
--	    let it = ["To: " ++ simpleQuote (showDestination cs kwds) , "--------", ""] 
--	    writeRawFile fn (stringToBytes $ unlines it)
--	    withProgram $ System.system (e ++ " +3 " ++ fn)
--	    pn <- fmap (lines . bytesToString)$ readRawFile fn 
--	    if (length pn > 1 && pn /= it) then do 
--		let (cs',kwds') = readDestination (drop 4 (head pn))
--		np <- puffTemplate
--		let p = np { cats = cs', fragments = fragments np ++ [("message/body",FragmentText (unlines $ (drop 2 pn)))] ++ [ ("message.keyword",FragmentText k) | k <- kwds']}
--		sendPuff p else 
--		    setMessage "Puff aborted."
